import json
import os
import shutil
import subprocess
from contextlib import contextmanager
from operator import attrgetter
from tempfile import NamedTemporaryFile
from typing import Iterator, List, Tuple

from match_video.anchor import Anchor


def write_anchors(
    input_video_path: str, output_video_path: str, anchors: List[Anchor]
) -> None:
    """Write anchors to a video file.

    Anchors are written as chapters in the video's metadata. Any existing anchors will
    be overwritten.

    Args:
        input_video_path: The path to a video.
        output_video_path: The path to write the video with anchors to.
        anchors: A list of anchors specifying the start of each half and any
            discontinuities in the video.
    """
    existing_metadata: str

    with NamedTemporaryFile("r") as existing_metadata_file:
        subprocess.run(
            [
                "ffmpeg",
                "-y",
                "-i",
                input_video_path,
                "-f",
                "ffmetadata",
                existing_metadata_file.name,
            ],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

        existing_metadata_file.seek(0)
        existing_metadata = existing_metadata_file.read()

    existing_metadata_without_chapters = (
        existing_metadata.split("[CHAPTER]")[0].strip()
        if "[CHAPTER]" in existing_metadata
        else existing_metadata
    )

    updated_metadata = existing_metadata_without_chapters

    for anchor in anchors:
        time = int(anchor.video_time * 1000)

        chapter = f"""

[CHAPTER]
TIMEBASE=1/1000
START={time}
END={time + 1}
title=Period {anchor.period}, {anchor.clock}"""

        updated_metadata += chapter

    with NamedTemporaryFile("w") as updated_metadata_file:
        updated_metadata_file.write(updated_metadata)
        updated_metadata_file.seek(0)

        def write_video_with_metadata(path: str) -> None:
            subprocess.run(
                [
                    "ffmpeg",
                    "-y",
                    "-i",
                    input_video_path,
                    "-i",
                    updated_metadata_file.name,
                    "-map_metadata",
                    "1",
                    "-codec",
                    "copy",
                    path,
                ],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )

        if os.path.exists(output_video_path) and os.path.samefile(
            input_video_path, output_video_path
        ):
            # ffmpeg can't update video metadata in place, so use an intermediate file

            with NamedTemporaryFile("w", suffix=".mp4") as intermediate_file:
                write_video_with_metadata(intermediate_file.name)
                shutil.copy2(intermediate_file.name, output_video_path)
        else:
            write_video_with_metadata(output_video_path)


def read_anchors(video_path: str) -> List[Anchor]:
    """Read the anchor points from the chapter metadata of a video file.

    Args:
        video_path: The path to a video.

    Returns:
        The list of anchors set for the video.

    Raises:
        ValueError: The video's metadata could not be read.
    """
    result_json = subprocess.run(
        [
            "ffprobe",
            "-v",
            "quiet",
            "-print_format",
            "json",
            "-show_error",
            "-show_chapters",
            video_path,
        ],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    result = json.loads(result_json.stdout)

    if "error" in result:
        raise ValueError(
            f"Unable to read the metadata for {video_path}, {result['error']['string']}"
        )

    def get_anchor(chapter: dict) -> Anchor:
        period_str, clock_str = chapter["tags"]["title"].split(", ")
        period = int(period_str.split("Period ")[-1])
        clock = float(clock_str)

        video_time = float(chapter["start_time"])

        return Anchor(period, clock, video_time)

    anchors = [get_anchor(chapter) for chapter in result["chapters"]]

    return anchors


def get_clip(
    video_path: str, period: int, start_clock: float, end_clock: float
) -> bytes:
    """Get a clip from a match by period and clock.

    Args:
        video_path: The path to a video.
        period: The period of the match the clip is in.
        start_clock: The start of the clip in seconds since the start of the period.
        end_clock: The end of the clip in seconds since the start of the period.

    Returns:
        The video clip as bytes.

    Raises:
        ValueError: The video does not have anchors or the clips is before the first
            anchor in its period.
    """
    anchors = read_anchors(video_path)

    if len(anchors) == 0:
        raise ValueError(f"{video_path} has no set anchors")

    start_video_time, end_video_time = _get_video_times(
        anchors, period, start_clock, end_clock
    )

    clip: bytes

    with _video_sans_chapters(video_path) as video_sans_chapters_path:
        with NamedTemporaryFile("rb", suffix=".mp4") as clip_file:
            _extract_clip(
                video_sans_chapters_path,
                clip_file.name,
                start_video_time,
                end_video_time,
            )

            clip = clip_file.read()

    return clip


def get_clips(video_path: str, clip_clocks: List[dict]) -> bytes:
    """Get clips from a match by period and clock.

    Args:
        video_path: The path to a video.
        clip_clocks: A list of clips to select and stitch together. Each clip
            dictionary should have a period, start_clock, and end_clock. These values
            are the same as with get_clip.

    Returns:
        The video clips as bytes.

    Raises:
        ValueError: The video does not have anchors or one of the clips is before the
            first anchor in its period.
    """
    anchors = read_anchors(video_path)

    if len(anchors) == 0:
        raise ValueError(f"{video_path} has no set anchors")

    clip_files: List

    with _video_sans_chapters(video_path) as video_sans_chapters_path:

        def create_clip_file(clip_info: dict):
            clip_file = NamedTemporaryFile("rb", suffix=".mp4")

            start_video_time, end_video_time = _get_video_times(
                anchors,
                clip_info["period"],
                clip_info["start_clock"],
                clip_info["end_clock"],
            )

            _extract_clip(
                video_sans_chapters_path,
                clip_file.name,
                start_video_time,
                end_video_time,
            )

            return clip_file

        clip_files = [create_clip_file(clip_info) for clip_info in clip_clocks]

    clips: bytes

    with NamedTemporaryFile("w") as clip_paths_file:
        clip_paths = "\n".join([f"file '{clip_file.name}'" for clip_file in clip_files])
        clip_paths_file.write(clip_paths)
        clip_paths_file.seek(0)

        with NamedTemporaryFile("rb", suffix=".mp4") as clips_file:
            subprocess.run(
                [
                    "ffmpeg",
                    "-y",
                    "-f",
                    "concat",
                    "-safe",
                    "0",
                    "-i",
                    clip_paths_file.name,
                    "-c",
                    "copy",
                    clips_file.name,
                ],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )

            clips = clips_file.read()

    for clip_file in clip_files:
        clip_file.close()

    return clips


def _get_video_times(
    anchors: List[Anchor], period: int, start_clock: float, end_clock: float
) -> Tuple[float, float]:
    """Convert period and match clocks into video times.

    Args:
        anchors: A video's anchors.
        period: The period of the match.
        start_clock: A time since the start of the period.
        end_clock: A time since the start of the period.

    Returns:
        A pair, (start_video_time, end_video_time), that correspond to start_clock and
        end_clock in video time.

    Raises:
        ValueError: There are no anchors in the period before start_clock.
    """
    prior_anchors = sorted(
        [
            anchor
            for anchor in anchors
            if anchor.period == period and anchor.clock <= start_clock
        ],
        key=attrgetter("clock"),
    )

    if len(prior_anchors) == 0:
        minute = int(start_clock / 60)
        second = int(start_clock % 60)

        raise ValueError(
            f"No anchors set in period {period} before {minute}:{second:02}"
        )

    last_anchor = prior_anchors[-1]
    start_offset = start_clock - last_anchor.clock
    start_video_time = last_anchor.video_time + start_offset

    duration = end_clock - start_clock
    end_video_time = start_video_time + duration

    return start_video_time, end_video_time


@contextmanager
def _video_sans_chapters(video_path: str) -> Iterator[str]:
    """Create a copy of a video without its chapters.

    Args:
        video_path: The path to a video.

    Yields:
        The path to the copied video file without chapters.
    """
    with NamedTemporaryFile("rb", suffix=".mp4") as video_sans_chapters_file:
        subprocess.run(
            [
                "ffmpeg",
                "-y",
                "-i",
                video_path,
                "-map",
                "0",
                "-vcodec",
                "copy",
                "-acodec",
                "copy",
                "-map_chapters",
                "-1",
                video_sans_chapters_file.name,
            ],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

        yield video_sans_chapters_file.name


def _extract_clip(
    input_video_path: str, output_video_path: str, start_time: float, end_time: float
) -> None:
    """Extract a clip from input_video_path and write it to output_video_path.

    Output can be incorrect if input_video_path has chapters. Use _video_sans_chapters
    to copy the video file without chapters before using this method.

    Args:
        input_video_path: The path to a video.
        output_video_path: The path to write the video with anchors to.
        start_time: The start of the clip in seconds since video start.
        end_time: The end of the clip in seconds since video start.
    """
    subprocess.run(
        [
            "ffmpeg",
            "-y",
            "-ss",
            f"{start_time:0.2f}",
            "-to",
            f"{end_time:0.2f}",
            "-i",
            input_video_path,
            "-acodec",
            "copy",
            "-vcodec",
            "copy",
            output_video_path,
        ],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )



from match_video.anchor import Anchor
from match_video.utils import get_clip, get_clips, read_anchors, write_anchors

__all__ = [
    "Anchor",
    "write_anchors",
    "read_anchors",
    "get_clip",
    "get_clips",
]



from typing import Optional

import typer

import match_video.utils as utils
from match_video.anchor import Anchor

app = typer.Typer()


@app.command()
def set_half_starts(
    input_video_path: str,
    first_half_start_time: str,
    second_half_start_time: str,
    output_video_path: Optional[str] = None,
) -> None:
    """Set the start times for each half of a match video.

    Args:
        input_video_path: The path to a video.
        first_half_start_time: The start of the first half as mm:ss in video time.
        second_half_start_time: The start of the second half as mm:ss in video time.
        output_video_path: The path to write the video with anchors to. Overwrite the
            input video if this is not specified.
    """
    if output_video_path is None:
        output_video_path = input_video_path

    def get_anchor(period: int, video_time_str: str) -> Anchor:
        video_minutes, video_seconds = video_time_str.split(":")
        video_time = int(video_minutes) * 60 + int(video_seconds)

        return Anchor(period, 0.0, video_time)

    anchors = [
        get_anchor(1, first_half_start_time),
        get_anchor(2, second_half_start_time),
    ]

    utils.write_anchors(input_video_path, output_video_path, anchors)


@app.command()
def read_anchors(video_path: str) -> None:
    """Read the anchors set for a video.

    Args:
        video_path: The path to a video.
    """
    anchors = utils.read_anchors(video_path)

    if len(anchors) == 0:
        typer.echo("No anchors set for video")

    for anchor in anchors:
        minute = int(anchor.clock / 60)
        second = int(anchor.clock % 60)

        video_minute = int(anchor.video_time / 60)
        video_second = int(anchor.video_time % 60)

        typer.echo(
            f"Period {anchor.period} {minute}:{second:02} | {video_minute}:{video_second:02} in video"
        )


if __name__ == "__main__":
    app()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['match_video']

package_data = \
{'': ['*']}

install_requires = \
['typer[all]>=0.4.0,<0.5.0']

extras_require = \
{'examples': ['streamlit>=1.12.0,<2.0.0',
              'jupyterlab>=3.6.1,<4.0.0',
              'xmltodict>=0.13.0,<0.14.0',
              'numpy==1.23.4',
              'pyarrow==10.0.0'],
 'lint': ['pre-commit>=2.19.0,<3.0.0',
          'black>=22.3.0,<23.0.0',
          'flake8>=4.0.1,<5.0.0',
          'isort>=5.6.4,<6.0.0'],
 'test': ['pytest>=7.1.2,<8.0.0',
          'coverage[toml]>=6.4,<7.0',
          'pytest-cov>=3.0.0,<4.0.0',
          'pytest-mock>=3.2.0,<4.0.0',
          'pytest-sugar>=0.9.4,<0.10.0']}

entry_points = \
{'console_scripts': ['match-video = match_video.cli:app']}

setup_kwargs = {
    'name': 'match-video',
    'version': '0.1.2',
    'description': 'A Python library that simplifies working with video from soccer matches.',
    'long_description': '# Match Video\n\nThis is a Python library that simplifies working with video from soccer matches. It allows match video to be selected intuitively by period number and clock, instead of absolute video time.\n\nTo accomplish this the start of each period is set as a chapter in the match video\'s metadata. Clips from the video can then be selected by period number and clock. ffmpeg handles both reading and writing the video chapter metadata and clip selection.\n\n## Installation\n\n### Requirements\n- Python 3.9 or newer\n- [ffmpeg](https://ffmpeg.org)\n\n```shell\npip install match-video\n```\n\n## Usage\n\nBefore the video can be used, the start time of each half needs to be set.\n\n```shell\nmatch-video set-half-starts path/to/video.mp4 0:04 63:20\n```\n\nThen it is easy to select match video by period and clock!\n\n```python\nimport match_video as mv\n\n# get the third minute of the match\nclip = mv.get_clip("path/to/video.mp4", period=1, start_clock=180, end_clock=240)\n\n# get the start of each half and concatenate them\nclip_clocks = [\n    {"period": 1, "start_clock": 0, "end_clock": 30},\n    {"period": 2, "start_clock": 0, "end_clock": 30},\n]\nclips = mv.get_clips("path/to/video.mp4", clip_clocks)\n```\n\nSee the [examples](https://gitlab.com/grantwenzinger/match-video/-/tree/main/examples) to see how to save or display video clips.\n\n## Support\n\n<grantwenzinger@gmail.com>\n\n## License\n\n[MIT](https://choosealicense.com/licenses/mit/)\n',
    'author': 'Grant Wenzinger',
    'author_email': 'grantwenzinger@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://gitlab.com/grantwenzinger/match-video',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.9.0,<4.0.0',
}


setup(**setup_kwargs)
